/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Ethernet webserver with reset push button
 *
 * Chip type           : Atmega168 or Atmega328 with ENC28J60
 * Note: there is a version number in the text. Search for version
 *********************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "avr_compat.h"
#include "net.h"

/* set output to Vcc, LED off */
#define LEDOFF PORTB|=(1<<PINB1)
/* set output to GND, LED on */
#define LEDON PORTB&=~(1<<PINB1)

// please modify the following two lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x11,0x00,0x27};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
static uint8_t myip[4] = {10,0,0,27};
// the default MAC to which you want to send a WOL (changable at runtime):
static uint8_t wolmac[6] = {0x00,0x14,0x51,0x00,0xa3,0x48};
//------------------- do not change below this line
static char *errmsg; // error text
// listen port for tcp/www (max range 1-254)
#define MYWWWPORT 80
//
#define BUFFER_SIZE 720
static uint8_t buf[BUFFER_SIZE+1];
// global string buffer
#define STR_BUFFER_SIZE 26
static char strbuf[STR_BUFFER_SIZE+1];
//
// the password string (only the first 7 char checked), (only a-z,0-9,_ characters):
static char password[10]="secret"; // must not be longer than 9 char

// 
uint8_t verify_password(char *str)
{
        // the first characters of the received string are
        // a simple password/cookie:
        if (strncmp(password,str,7)==0){
                return(1);
        }
        return(0);
}
// search for a string of the form key=value in
// a string that looks like q?xyz=abc&uvw=defgh HTTP/1.1\r\n
//
// The returned value is stored in the global var strbuf
uint8_t find_key_val(char *str,char *key)
{
        uint8_t found=0;
        uint8_t i=0;
        char *kp;
        kp=key;
        while(*str &&  *str!=' ' && found==0){
                if (*str == *kp){
                        kp++;
                        if (*kp == '\0'){
                                str++;
                                kp=key;
                                if (*str == '='){
                                        found=1;
                                }
                        }
                }else{
                        kp=key;
                }
                str++;
        }
        if (found==1){
                // copy the value to a buffer and terminate it with '\0'
                while(*str &&  *str!=' ' && *str!='&' && i<STR_BUFFER_SIZE){
                        strbuf[i]=*str;
                        i++;
                        str++;
                }
                strbuf[i]='\0';
        }
        // return the length of the value
        return(i);
}

// convert a single hex digit character to its integer value
unsigned char h2int(char c)
{
        if (c >= '0' && c <='9'){
                return((unsigned char)c - '0');
        }
        if (c >= 'a' && c <='f'){
                return((unsigned char)c - 'a' + 10);
        }
        if (c >= 'A' && c <='F'){
                return((unsigned char)c - 'A' + 10);
        }
        return(0);
}

// decode a url string e.g "hello%20joe" or "hello+joe" becomes "hello joe"
void urldecode(char *urlbuf)
{
        char c;
        char *dst;
        dst=urlbuf;
        while ((c = *urlbuf)) {
                if (c == '+') c = ' ';
                if (c == '%') {
                        urlbuf++;
                        c = *urlbuf;
                        urlbuf++;
                        c = (h2int(c) << 4) | h2int(*urlbuf);
                }
                *dst = c;
                dst++;
                urlbuf++;
        }
        *dst = '\0';
}

// parse a string and extract the MAC to bytestr
// return 0 on success
uint8_t parse_mac(uint8_t *bytestr,char *str)
{
        char *sptr;
        uint8_t i=0;
        sptr=NULL;
        while(i<6){
                bytestr[i]=0;
                i++;
        }
        i=0;
        while(*str && i<6){
                // if a number or letter then start
                if (sptr==NULL && isalnum(*str)){
                        sptr=str;
                }
                if (*str == ':'){
                        *str ='\0';
                        bytestr[i]=(strtol(sptr,(char **)NULL,16)&0xff);
                        i++;
                        sptr=NULL;
                }
                str++;
        }
        *str ='\0';
        if (i==5){
                bytestr[i]=(strtol(sptr,(char **)NULL,16)&0xff);
                return(0);
        }
        return(1);
}


// parse a string and extract the IP to bytestr
// return 0 on success
uint8_t parse_ip(uint8_t *bytestr,char *str)
{
        char *sptr;
        uint8_t i=0;
        sptr=NULL;
        while(i<4){
                bytestr[i]=0;
                i++;
        }
        i=0;
        while(*str && i<4){
                // if a number then start
                if (sptr==NULL && isdigit(*str)){
                        sptr=str;
                }
                if (*str == '.'){
                        *str ='\0';
                        bytestr[i]=(atoi(sptr)&0xff);
                        i++;
                        sptr=NULL;
                }
                str++;
        }
        *str ='\0';
        if (i==3){
                bytestr[i]=(atoi(sptr)&0xff);
                return(0);
        }
        return(1);
}

// take a byte string and display it  (base is 10 for ip and 16 for mac addr)
void mk_net_str(char *resultstr,uint8_t *bytestr,uint8_t len,char separator,uint8_t base)
{
        uint8_t i=0;
        uint8_t j=0;
        while(i<len){
                itoa((int)bytestr[i],&resultstr[j],base);
                // search end of str:
                while(resultstr[j]){j++;}
                resultstr[j]=separator;
                j++;
                i++;
        }
        j--;
        resultstr[j]='\0';
}

// takes a string of the form ack?pw=xxx&rst=1 and analyse it
// return values:  0 error
//                 1 /wol page
//                 2 /ipc page
//                 3 /cnf page 
uint8_t analyse_get_url(char *str)
{
        errmsg="invalid pw";
        if (strncmp("cnf",str,3)==0){
                return(3);
        }
        // actions reset stop start
        if (strncmp("wol",str,3)==0){
                if (find_key_val(str,"pw")){
                        urldecode(strbuf);
                        if (verify_password(strbuf)){
                                if (find_key_val(str,"mac")){
                                        urldecode(strbuf);
                                        if (parse_mac(wolmac,strbuf)){
                                                errmsg="invalid mac";
                                                return(0);
                                        }
                                        return(1);
                                }
                        }
                }
                return(0);
        }
        // change own ip and pw
        if (strncmp("ipc",str,3)==0){
                if (find_key_val(str,"pw")){
                        urldecode(strbuf);
                        if (verify_password(strbuf)){
                                if (find_key_val(str,"nip")){
                                        urldecode(strbuf);
                                        if (parse_ip(myip,strbuf)){
                                                errmsg="invalid ip";
                                                return(0);
                                        }
                                }
                                if (find_key_val(str,"npw")){
                                        urldecode(strbuf);
                                        strbuf[7]='\0';
                                        strcpy(password,strbuf);
                                        return(2);
                                }
                        }
                }
                return(0);
        }
        errmsg="inv. url";
        return(0);
}


uint16_t http200ok(void)
{
        return(fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n")));
}


// prepare the webpage by writing the data to the tcp send buffer
uint16_t print_webpage(uint8_t *buf)
{
        uint16_t plen;
        plen=http200ok();
        plen=fill_tcp_data_p(buf,plen,PSTR("Menu: <a href=/cnf>config</a> "));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Send Wake on Lan</h2>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<pre><form action=/wol method=get>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("MAC: <input type=text size=20 name=mac value="));
        mk_net_str(strbuf,wolmac,6,':',16);
        plen=fill_tcp_data(buf,plen,strbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR(">\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR(" pw: <input type=password size=8 name=pw>\n<input type=submit value=\"send WOL\"></form>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\n<hr>Version 0.1"));

        return(plen);
}

uint16_t print_webpage_config(uint8_t *buf)
{
        uint16_t plen;
        plen=http200ok();
        plen=fill_tcp_data_p(buf,plen,PSTR("Menu: <a href=/>main menu</a> "));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Change IP/password</h2><pre>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<form action=/ipc method=get>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("old pw: <input type=password size=8 name=pw>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("new pw: <input type=password size=8 name=npw>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("new IP: <input type=text size=12 name=nip value="));
        mk_net_str(strbuf,myip,4,'.',10);
        plen=fill_tcp_data(buf,plen,strbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR("><br>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<input type=submit value=\"change\"></form><hr>"));
        return(plen);
}


uint16_t print_webpage_confirm(uint8_t *buf)
{
        uint16_t plen;
        plen=http200ok();
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>OK</h2> <a href=\"/\">-&gt;continue</a></p>\n"));
        return(plen);
}


void store_in_eeprom(void)
{
        eeprom_write_byte((uint8_t *)0x0,19); // magic number
        eeprom_write_block((uint8_t *)myip,(void *)1,sizeof(myip));
        password[7]='\0'; // make sure it is terminated, should not be necessary
        eeprom_write_block((char *)password,(void *)6,sizeof(password));
}

int main(void){

        uint16_t plen;
        uint16_t dat_p;
        uint8_t cmd;
        uint8_t sendwol=0;
        
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_loop_1(50); // 12ms

        /* enable PD2/INT0, as input */
        DDRD&= ~(1<<DDD2);

        // test button
        cbi(DDRD,PIND6);
        sbi(PORTD,PIND6); // internal pullup resistor on
        _delay_loop_1(50); // 12ms
        // read eeprom values unless, the PD6 pin is connected to GND during bootup
        // PD6 is used to reset to factory default. Note that factory default is
        // not stored back to eeprom. 
        if (eeprom_read_byte((uint8_t *)0x0) == 19 && ! bit_is_clear(PIND,PIND6)){
                // ok magic number matches accept values
                eeprom_read_block((uint8_t *)myip,(void *)1,sizeof(myip));
                eeprom_read_block((char *)password,(void *)6,sizeof(password));
                password[7]='\0'; // make sure it is terminated, should not be necessary
        }

        /*initialize enc28j60*/
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_loop_1(50); // 12ms
        
        // LED
        /* enable PB1, LED as output */
        //DDRB|= (1<<DDB1);
        /* set output to Vcc, LED off */
        //LEDOFF;

        // the transistor on PD7
        DDRD|= (1<<DDD7);
        PORTD &= ~(1<<PIND7);// transistor off
        
        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        _delay_loop_1(50); // 12ms
        

        //init the ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,MYWWWPORT);

        sei(); // interrupt enable

        while(1){
                // spontanious messages must not interfer with
                // web pages
                if (sendwol){
                        sendwol=0;
                        send_wol(buf,wolmac);
                }
                // get the next new packet:
                plen = enc28j60PacketReceive(BUFFER_SIZE, buf);

                /*plen will be unequal to zero if there is a valid 
                 * packet (without crc error) */
                if(plen==0){
                        continue;
                }
                        
                // arp is broadcast if unknown but a host may also
                // verify the mac address by sending it to
                // a unicast address.
                if(eth_type_is_arp_and_my_ip(buf,plen)){
                        if (eth_type_is_arp_req(buf)){
                                make_arp_answer_from_request(buf);
                        }
                        continue;
                }

                // check if ip packets are for us:
                if(eth_type_is_ip_and_my_ip(buf,plen)==0){
                        continue;
                }
                
                if(buf[IP_PROTO_P]==IP_PROTO_ICMP_V && buf[ICMP_TYPE_P]==ICMP_TYPE_ECHOREQUEST_V){
                        // a ping packet, let's send pong
                        make_echo_reply_from_request(buf,plen);
                        continue;
                }
                // tcp port www start, compare only the lower byte
                if (buf[IP_PROTO_P]==IP_PROTO_TCP_V&&buf[TCP_DST_PORT_H_P]==0&&buf[TCP_DST_PORT_L_P]==MYWWWPORT){
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_SYN_V){
                                make_tcp_synack_from_syn(buf);
                                // make_tcp_synack_from_syn does already send the syn,ack
                                continue;
                        }
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_ACK_V){
                                init_len_info(buf); // init some data structures
                                // we can possibly have no data, just ack:
                                dat_p=get_tcp_data_pointer();
                                if (dat_p==0){
                                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_FIN_V){
                                                // finack, answer with ack
                                                make_tcp_ack_from_any(buf);
                                        }
                                        // just an ack with no data, wait for next packet
                                        continue;
                                }
                                if (strncmp("GET ",(char *)&(buf[dat_p]),4)!=0){
                                        // head, post and other methods:
                                        //
                                        // for possible status codes see:
                                        // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<h1>200 OK</h1>"));
                                        goto SENDTCP;
                                }
                                if (strncmp("/ ",(char *)&(buf[dat_p+4]),2)==0){
                                        plen=print_webpage(buf);
                                        goto SENDTCP;
                                }
                                cmd=analyse_get_url((char *)&(buf[dat_p+5]));
                                //
                                // error, default, will get overwritte in case 
                                // something else is selected:
                                plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<h1>Error</h1>"));
                                plen=fill_tcp_data(buf,plen,errmsg);
                                plen=fill_tcp_data_p(buf,plen,PSTR("<br><br><a href=/>-&gt;continue</a>\n"));
                                // for possible status codes see:
                                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                                if (cmd==1){
                                        sendwol=1;
                                        plen=print_webpage_confirm(buf);
                                }
                                if (cmd==2){
                                        store_in_eeprom();
                                        plen=print_webpage_confirm(buf);
                                }
                                if (cmd==3){
                                        plen=print_webpage_config(buf);
                                }
                                //
SENDTCP:
                                make_tcp_ack_from_any(buf); // send ack for http get
                                make_tcp_ack_with_data(buf,plen); // send data
                                continue;
                        }

                }
                // tcp port www end
                //
                // udp not implemented
        }
        return (0);
}
